﻿#pragma TextEncoding="UTF-8"
#pragma rtGlobals=3
#pragma moduleName=XYbrowser
#pragma version=1.20
#include <Resize Controls>

// A browser gui for previewing y vs x plots of paired waves.
// Waves are paired by specifying the required X wave in the
// wavenote of the Y wave, in the form "Xwave:wavename".

// The browser is useful for building Y vs X plots of many pairs of
// waves. X-Y wave pairs are used rather than 2D waves for compatibility
// with various tools for spectroscopy that will not work with 2D waves.

// This is designed to be combined with a custom file loader
// that sets the wavenote. An example is the WiRE loader for
// Renishaw Raman files, https://www.wavemetrics.com/node/21301

// set this constant to format graphs for Raman spectroscopy
static constant kRaman=1

// headers used by the updater project
static constant kProjectID=21303 // the project node on IgorExchange
static strconstant ksShortTitle="XYBrowser" // the project short title on IgorExchange

menu "Data"
	submenu "Packages"
		"X-Y Browser", /Q, XYbrowser#MakeXYBrowserPanel()
	end
end
	
// returns text wave containing names of waves in dfr with
// x-wave specified in wavenote (Xwave:wavename). X-wave must
// exist in dfr.
static function /wave XYBrowserWaves(dfr)
	DFREF dfr
	
	string s_allwaves, s_out="", s_note, s_item, s_x
	variable i, numWaves
	
	s_allwaves=WaveListDFR("*",";","DIMS:1", dfr)
	numWaves=ItemsInList(s_allwaves)
	
	for(i=0;i<numWaves;i+=1)
		s_item=StringFromList(i,s_allwaves)
		wave /Z w_x=XYBrowserXWaveRef(dfr:$s_item)
		if(WaveExists(w_x))
			s_out=AddListItem(s_item, s_out)
		endif
	endfor
	return ListToTextWave(s_out, ";")
end

static function /S WaveListDFR(matchStr, separatorStr, optionsStr, dfr)
	string matchStr, separatorStr, optionsStr; DFREF dfr
	
	string s
	variable err
	DFREF saveDFR = GetDataFolderDFR()
	SetDataFolder dfr
	
	try
		err = GetRTError(1)
		s = WaveList(matchStr, separatorStr, optionsStr); AbortOnRTE
	catch
		err = GetRTError(1)
		SetDataFolder saveDFR
		return ""
	endtry
	SetDataFolder saveDFR
	return s
end

// returns waveref for the X wave paired with wave w
static function /WAVE XYBrowserXWaveRef(w)
	wave w
	
	if(WaveExists(w)==0)
		return $""
	endif
	DFREF dfr=GetWavesDataFolderDFR(w)
	string s_x=StringByKey("Xwave", note(w), ":", "\r")
	if(cmpstr(s_x[0]," ")==0)
		do // remove whitespace after key separator
			s_x=ReplaceString(" ", s_x, "", 0, 1)
		while((cmpstr(s_x[0]," ")==0)&&strlen(s_x))
	endif
	do
		s_x=RemoveEnding(s_x, " ")
	while(cmpstr(s_x[strlen(s_x)-1]," ")==0)
	
	return dfr:$s_x
end

static function MakeXYBrowserPanel()
	
	DoWindow /F XYBrowserPanel
	if(v_flag)
		XYbrowser#RefreshXYBrowser(0)
		return 1
	endif
	
	NewDataFolder /O root:Packages
	NewDataFolder /O root:Packages:XYBrowser
	DFREF dfr=root:Packages:XYBrowser
	
	Make /o/T/n=0 dfr:XYWaves /wave=XYWaves
	Make /o/n=0 dfr:selWave /wave=selWave
	
	NewPanel /K=1/N=XYBrowserPanel/W=(100,100,840,470) as "X-Y Browser"
	
	DefineGuide/W=XYBrowserPanel guideR={FR,-20}
	DefineGuide/W=XYBrowserPanel guideDiv={FL,0.3,Fr}
	DefineGuide/W=XYBrowserPanel guideL={guideDiv,10}
	DefineGuide/W=XYBrowserPanel guideT={FT,20}
	DefineGuide/W=XYBrowserPanel guideB={FB,-45}
	
	Display/FG=(guideL,guideT,guideR,guideB)/HOST=XYBrowserPanel /N=PreviewPlot
	ModifyGraph /W=XYBrowserPanel#PreviewPlot frameStyle=5
	SetActiveSubwindow XYBrowserPanel
	
	ListBox ListBoxDataWaves,pos={20, 20},size={192, 305}, mode=9, listwave=XYWaves, selWave=selWave
	ListBox ListBoxDataWaves, proc=XYbrowser#XYBrowserListBoxProc
	
	Button ButtonBrowseDF,win=XYBrowserPanel, pos={20,337},size={25,20}, title="..."
	Button ButtonBrowseDF,win=XYBrowserPanel, proc=XYbrowser#XYBrowserButtonProc
	Button ButtonBrowseDF,win=XYBrowserPanel, help={"Click to select data folder"}
	Button ButtonBrowseDF,win=XYBrowserPanel, userdata(DF)=GetDataFolder(1)
	TitleBox DFbox,win=XYBrowserPanel, pos={50, 342},size={150,25},frame=0
	TitleBox DFbox,win=XYBrowserPanel, title="\\Z10\\f01Data Folder \\f00"+GetDataFolder(1)
	
	Button ButtonNewGraph,win=XYBrowserPanel, pos={640,337},size={80,25},title="New Graph"
	Button ButtonNewGraph,win=XYBrowserPanel, disable=2,proc=XYbrowser#XYBrowserButtonProc
	Button ButtonNewGraph,win=XYBrowserPanel, help={"Display the selected X-Y pairs in a graph window"}
	Button ButtonAppendToTop,win=XYBrowserPanel, pos={510,337},size={100,25},title="Append to top"
	Button ButtonAppendToTop,win=XYBrowserPanel, disable=2, proc=XYbrowser#XYBrowserButtonProc
	Button ButtonAppendToTop,win=XYBrowserPanel, help={"Add the selected X-Y pairs to the top graph window"}
	
	XYbrowser#RefreshXYBrowser(1) // update listbox waves
	SetWindow XYBrowserPanel, activeChildFrame=0
	SetWindow XYBrowserPanel, hook(RBhook)=XYbrowser#XYBrowserHook
	
	// resizing user data for controls
	ListBox ListBoxDataWaves,win=XYBrowserPanel,userdata(ResizeControlsInfo)= A"!!,BY!!#<X!!#AO!!#BRJ,fQL!!#](Aon\"Qzzzzzzzzzzzzzz!!#N3Bk1ctB6@p!AOL6U"
	ListBox ListBoxDataWaves,win=XYBrowserPanel,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzz!!#u:Du]k<zzzzzzzzzzz"
	ListBox ListBoxDataWaves,win=XYBrowserPanel,userdata(ResizeControlsInfo) += A"zzz!!#N3Bk1ctB6@p!AO6@szzzzzzzzzzzz!!!"
	TitleBox DFbox,win=XYBrowserPanel,userdata(ResizeControlsInfo)= A"!!,DW!!#Be!!#?c!!#;]z!!#](Aon\"Qzzzzzzzzzzzzzz!!#](Aon\"Qzz"
	TitleBox DFbox,win=XYBrowserPanel,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	TitleBox DFbox,win=XYBrowserPanel,userdata(ResizeControlsInfo) += A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	Button ButtonAppendToTop,win=XYBrowserPanel,userdata(ResizeControlsInfo)= A"!!,Ie!!#BbJ,hpW!!#=+z!!#o2B4uAezzzzzzzzzzzzzz!!#o2B4uAezz"
	Button ButtonAppendToTop,win=XYBrowserPanel,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	Button ButtonAppendToTop,win=XYBrowserPanel,userdata(ResizeControlsInfo) += A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	Button ButtonNewGraph,win=XYBrowserPanel,userdata(ResizeControlsInfo)= A"!!,J1!!#BbJ,hp/!!#=+z!!#o2B4uAezzzzzzzzzzzzzz!!#o2B4uAezz"
	Button ButtonNewGraph,win=XYBrowserPanel,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	Button ButtonNewGraph,win=XYBrowserPanel,userdata(ResizeControlsInfo) += A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"
	Button ButtonBrowseDF,win=XYBrowserPanel,userdata(ResizeControlsInfo)= A"!!,BY!!#BbJ,hmV!!#<Xz!!#](Aon\"Qzzzzzzzzzzzzzz!!#](Aon\"Qzz"
	Button ButtonBrowseDF,win=XYBrowserPanel,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzz!!#?(FEDG<zzzzzzzzzzz"
	Button ButtonBrowseDF,win=XYBrowserPanel,userdata(ResizeControlsInfo) += A"zzz!!#?(FEDG<zzzzzzzzzzzzzz!!!"

	// resizing user data for panel window
	SetWindow XYBrowserPanel,userdata(ResizeControlsInfo)= A"!!*'\"z!!#DI!!#Bszzzzzzzzzzzzzzzzzzzzz"
	SetWindow XYBrowserPanel,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzzzzzzzzzzzzzzz"
	SetWindow XYBrowserPanel,userdata(ResizeControlsInfo) += A"zzzzzzzzzzzzzzzzzzz!!!"
	SetWindow XYBrowserPanel,userdata(ResizeControlsGuides)=  "guideR;guideDiv;guideL;guideT;guideB;"
	SetWindow XYBrowserPanel,userdata(ResizeControlsInfoguideR)=  "NAME:guideR;WIN:XYBrowserPanel;TYPE:User;HORIZONTAL:0;POSITION:720.00;GUIDE1:FR;GUIDE2:;RELPOSITION:-20;"
	SetWindow XYBrowserPanel,userdata(ResizeControlsInfoguideDiv)=  "NAME:guideDiv;WIN:XYBrowserPanel;TYPE:User;HORIZONTAL:0;POSITION:222.00;GUIDE1:FL;GUIDE2:Fr;RELPOSITION:0.3;"
	SetWindow XYBrowserPanel,userdata(ResizeControlsInfoguideL)=  "NAME:guideL;WIN:XYBrowserPanel;TYPE:User;HORIZONTAL:0;POSITION:232.00;GUIDE1:guideDiv;GUIDE2:;RELPOSITION:10;"
	SetWindow XYBrowserPanel,userdata(ResizeControlsInfoguideT)=  "NAME:guideT;WIN:XYBrowserPanel;TYPE:User;HORIZONTAL:1;POSITION:20.00;GUIDE1:FT;GUIDE2:;RELPOSITION:20;"
	SetWindow XYBrowserPanel,userdata(ResizeControlsInfoguideB)=  "NAME:guideB;WIN:XYBrowserPanel;TYPE:User;HORIZONTAL:1;POSITION:325.00;GUIDE1:FB;GUIDE2:;RELPOSITION:-45;"
	
	// resizing panel hook
	SetWindow XYBrowserPanel hook(ResizeControls)=ResizeControls#ResizeControlsHook
	
	return 1
end

static Function XYBrowserHook(s)
	STRUCT WMWinHookStruct &s
	
	if(s.eventCode==0) // activate
		refreshXYBrowser(0)
	endif
	return 0
end

// updates plot and buttons
static function XYBrowserListBoxProc(s) : ListboxControl
	STRUCT WMListboxAction &s
	
	if(s.eventCode<4 || s.eventCode>5)
		return 0
	endif
	
	if(WinType("XYBrowserPanel#PreviewPlot")==0)
		return 0
	endif
	
	string s_topgraph=WinList("*", ";", "WIN:1")
	
	Button buttonNewGraph,win=XYBrowserPanel, disable=2*(sum(s.selWave)==0)
	Button buttonAppendToTop,win=XYBrowserPanel, disable=2*((sum(s.selWave)==0) || (strlen(s_topgraph)==0))
	
	string plotList=TraceNameList("XYBrowserPanel#PreviewPlot",";",1)
	variable isPlotted
	DFREF wavDFR = $GetUserData("XYBrowserPanel","ButtonBrowseDF", "DF")

	variable i, imax=numpnts(s.selWave)
	for(i=0;i<imax;i+=1)
		isPlotted=(FindListItem(s.ListWave[i], plotList)>-1)
		if(s.selWave[i])
			if(isPlotted==0)
				wave w=wavDFR:$s.ListWave[i]
				wave w_x=XYBrowserXWaveRef(w)
				if(WaveExists(w) && WaveExists(w_x))
					AppendToGraph /W=XYBrowserPanel#PreviewPlot w vs w_x
				endif
			else
				plotList=RemoveFromList(s.ListWave[i],plotList)
			endif
		endif
	endfor
	
	// remove other waves from plot
	imax=ItemsInList(plotList)
	for(i=0;i<imax;i+=1)
		RemoveFromGraph /Z/W=XYBrowserPanel#PreviewPlot $StringFromList(i, plotList)
	endfor
	
	SetAxis /Z/A/W=XYBrowserPanel#PreviewPlot
	return 0
end

static function XYBrowserButtonProc(s) : buttonControl
	STRUCT WMbuttonAction &s
	
	variable hist=1 // choose whether to record graph building commands to history
	
	if(s.eventCode!=2)
		return 0
	endif
	
	DFREF dfr=root:Packages:XYBrowser
	wave /T XYWaves=dfr:XYWaves
	wave selWave=dfr:selWave
		
	if(stringmatch(s.ctrlName, "buttonBrowseDF"))
		XYbrowser#makeDFbrowser()
		return 0
	endif
	
	if(stringmatch(s.ctrlName, "buttonNewGraph"))
		Display
		if (hist)
			Print "•Display"
		endif
	endif
	
	DFREF wavDFR = $GetUserData("XYBrowserPanel","ButtonBrowseDF", "DF")
	variable i, imax=numpnts(selWave)
	string cmd=""
	
	for(i=0;i<imax;i+=1)
		if (selWave[i])
			wave w=wavDFR:$XYWaves[i]
			if (WaveExists(w)==0)
				// shouldn't arrive here!
				Print "could not find "+XYWaves[i]
				continue
			endif
			wave w_x=XYBrowserXWaveRef(w)
			if (WaveExists(w_x)==0)
				Print "could not find x wave for "+XYWaves[i]
				continue
			endif
			AppendToGraph w vs w_x
			if (hist) // record the graph building commands to history
				printf "•AppendToGraph %s vs %s\r", GetWavesDataFolder(w,4), GetWavesDataFolder(w_x,4)
			endif
		endif
	endfor
	
	if(kRaman && stringmatch(s.ctrlName, "buttonNewGraph"))
		Label/Z left "Intensity"
		Label/Z bottom "Wavenumber (cm\\S-1\\M)"
	endif
	return 0
end

static function MakeDFbrowser()
	
	CreateBrowser/M prompt="Switch data folder:"
	ModifyBrowser/M showWaves=0, showVars=0, showStrs=0, ShowInfo=0, showPlot=0
	ModifyBrowser/M expandAll, clearSelection
	ModifyBrowser/M selectList=GetDataFolder(1, $GetUserData("XYBrowserPanel","ButtonBrowseDF", "DF"))+";"
	ModifyBrowser/M showModalBrowser
	
	if (V_Flag == 0)
		return 0 // User cancelled
	endif
	
	if(ItemsInList(s_browserlist)!=1)
		return 0
	endif
	s_browserlist=StringFromList(0,s_browserlist)
	if(DataFolderExists(S_BrowserList))
		TitleBox DFbox win=XYBrowserPanel, title="\\Z10\\f01Data Folder \\f00"+S_BrowserList
		Button ButtonBrowseDF,win=XYBrowserPanel, userdata(DF)=S_BrowserList
	endif
	XYbrowser#refreshXYBrowser(1)
	return 1
end

static function refreshXYBrowser(resetSelection)
	variable resetSelection
	
	DFREF dfr=root:Packages:XYBrowser
	wave /T XYWaves=dfr:XYWaves
	wave selWave=dfr:selWave
	
	DFREF wavDFR = $GetUserData("XYBrowserPanel","ButtonBrowseDF", "DF")
	if (DataFolderRefStatus(wavDFR) == 0)
		TitleBox DFbox win=XYBrowserPanel, title="\\Z10\\f01Data Folder \\f00"+GetDataFolder(1)
		Button ButtonBrowseDF,win=XYBrowserPanel, userdata(DF)=GetDataFolder(1)
		DFREF wavDFR = GetDataFolderDFR()
		resetSelection=1
	endif

	variable i=numpnts(XYWaves)
	wave /t w=XYBrowserWaves(wavDFR) // get a new list of waves
	Duplicate /o w dfr:XYWaves
	if (numpnts(XYWaves)!=i || resetSelection)
		Make /o/n=(numpnts(XYWaves)) dfr:selWave=0
		wave selWave=dfr:selWave
	endif
		
	// trigger XYBrowserListBoxProc to update preview plot and button status
	STRUCT WMListboxAction s
	s.eventCode=4
	wave s.selWave=selWave
	wave /T s.listWave=XYWaves
	XYBrowserListBoxProc(s)
	
	return 1
end

// Provides a quick way to pair x and y waves to test the X-Y Browser
// Right-click on a trace in an X-Y plot and select "Pair X-wave as
// plotted." If both waves are in the same data folder the pair should now
// be selectable in the X-Y browser panel.
menu "TracePopup", dynamic
	XYbrowser#XYTraceMenu("Pair X-wave as plotted"), /Q,  XYbrowser#PairTrace()
end

static function PairTrace()
	
	// figure out graph and trace names
	GetLastUserMenuInfo  // sets s_tracename, s_graphname
	wave w_x=XWaveRefFromTrace(S_graphName,S_traceName)
	wave w=TraceNameToWaveRef(S_graphName,S_traceName)
	note /K w, ReplaceStringByKey("Xwave", note(w), NameOfWave(w_x), ":", "\r")
	return 1
end

static function /s XYTraceMenu(s)
	string s
	
	if (WinType("")!=1)
		return "" // don't do anything if Igor is just rebuilding the menu
	endif
	
	// figure out graph and trace names
	GetLastUserMenuInfo
	return SelectString(WaveExists(XWaveRefFromTrace(s_graphname, s_tracename)), "", s)
end